#!/usr/bin/env ruby -w

# Implementation of Fast Walsh Transforms (FWTs) using sequency and
# Hadamard in-place transforms.  Both algorithms are O(n log(n)).
#
# Note that these transforms are their own inverse, to within a scale
# factor of array.length, because the transform matrix is orthogonal and
# symmetric about its diagonal.

class Array

  # Perform a fast Walsh transformation using a Manz sequency ordered
  # in-place algorithm. The array is modified by this algorithm.
  def sequency
    raise "#{self.class}.#{__method__}: array length must be a power of 2" \
      unless length.power_of_2?
    j = 0
    0.upto(length - 3) do |i|  # Bit reversal sorting
      self[i], self[j] = self[j], self[i] if (i < j)
      k = length >> 1
      while (k <= j)
        j -= k
        k >>= 1
      end
      j += k
    end
    offset = length
    while(offset > 1)
      lag = offset >> 1
      ngroups = length / offset
      ngroups.times do |group|
        lag.times do |i|
          j = i + group * offset
          k = j + lag
          if (group.odd?)
            self[j], self[k] = self[j] - self[k], self[j] + self[k]
          else
            self[j], self[k] = self[j] + self[k], self[j] - self[k]
          end
        end
      end
      offset = lag
    end
    self
  end

  # Perform a fast Walsh transformation using a Hadamard (natural) ordered
  # in-place algorithm. The array is modified by this algorithm.
  def hadamard
    raise "#{self.class}.#{__method__}: array length must be a power of 2" \
      unless length.power_of_2?
    lag = 1
    while lag < length
      offset = lag << 1
      ngroups = length / offset
      ngroups.times do |group|
        lag.times do |base|
          j = base + group * offset
          k = j + lag
          self[j], self[k] = self[j] + self[k], self[j] - self[k]
        end
      end
      lag = offset
    end
    self
  end

end

class Integer
  # Boolean test of whether an Integer is a pure power of two.
  # This is an O(1) algorithm.
  def power_of_2?
    (self.nonzero? & (self-1)) == 0
  end
end
